unit OmegaMusic;

interface

uses
  SysUtils,
  Classes,
  Windows,
  Forms,
  Messages,
  Math,
  DirectShow,
  ActiveX,
  OmegaUtils;

type
  TSongListItem = class(TCollectionItem)
  private
    FPan: Integer;
    FVolume: Integer;
    FPosition: Double;
    FLoaded: Boolean;
    FPaused: Boolean;
    FInitialized: Boolean;
    FFileName: String;
    FName: String;
    FLoop: Boolean;

    FGraphBuilder: IGraphBuilder;
    FMediaPosition: IMediaPosition;
    FMediaControl: IMediaControl;
    FBasicAudio: IBasicAudio;
  protected
    procedure CleanUpDirectShow;
    procedure InitDirectShow;     
    procedure SetPan(Value: Integer);
    procedure SetVolume(Value: Integer);
    procedure SetPosition(Value: Double);
    function GetPosition: Double;
    function GetDuration: Double;
    procedure LoopIt;
    function GetDisplayName: String; override;
  public
    property Position: Double read GetPosition write SetPosition;
    property Duration: Double read GetDuration;
    constructor Create(Collection: TCollection); override;
    destructor Destroy; override;
    procedure Load; overload;
    procedure Load(const FileName: String); overload;
    function IsPlaying: Boolean;
    procedure Play;
    procedure Stop;
    procedure Pause;
    procedure Init;
  published
    property Volume: Integer read FVolume write SetVolume default 100;
    property Pan: Integer read FPan write SetPan default 0;
    property FileName: String read FFileName write FFilename;
    property Name: String read FName write FName;
    property Loop: Boolean read FLoop write FLoop;
  end;

  TSongListCollection = class(TOwnedCollection)
  private
    procedure Loop;
  protected
    function GetItem(Index: Integer): TSongListItem;
    procedure SetItem(Index: Integer; Value: TSongListItem);
    procedure Update(Item: TCollectionItem); override;
  public
    constructor Create(AOwner: TComponent);
    destructor Destroy; override;
    function Add: TSongListItem;
    function Insert(Index: Integer): TSongListItem;
    function Find(AName: String): TSongListItem;
    //procedure SaveToFile(const FileName: String);
    //procedure SaveToStream(Stream: TStream);
    //procedure LoadFromFile(const FileName: String);
    //procedure LoadFromStream(Stream: TStream);
    
    property Items[Index: Integer]: TSongListItem read GetItem write SetItem;
  published
  end;

  TOmegaMusic = class(TComponent)
  private
    FSongList: TSongListCollection;
    procedure SetSongList(const Value: TSongListCollection);
  protected
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    procedure Init;
    procedure Loop;
  published
    property Songs: TSongListCollection read FSongList write SetSongList;
  end;

implementation

//-----------------------------------------------------

function DSVolume(Value: Integer): Integer;
var
  iNew, iVol: Integer;
  fPercent, fLog: Extended;
begin
  iNew := Value;
  iVol :=  -10000;

  if iNew > 0 then
  begin
    fPercent := iNew*0.01;
    fLog := Log10(fPercent);
    iVol := RoundNormal(fLog * 1000);
  end;

  Result := iVol;
end;

//-----------------------------------------------------

function DSPan(Value: Integer): Integer;
var
  iNew, iPan, Temp: Integer;
  fPercent, fLog: Extended;
begin
  iNew := Value;
  iPan := 10000;

  if (iNew > 0)  then
  begin
    if (iNew < 100) then
    begin
      Temp := abs(iNew-100);
      fPercent := Temp*0.01;
      fLog := Log10(fPercent);
      iPan := -RoundNormal(fLog * 1000);
    end;
    Result := iPan;
    Exit;
  end;

  iPan := -10000;

  if iNew < 0 then
  begin
    if iNew > -100 then
    begin
      Temp := abs(iNew+100);
      fPercent := Temp*0.01;
      fLog := Log10(fPercent);
      iPan := RoundNormal(fLog * 1000);
    end;
    Result := iPan;
    Exit;
  end;

  iPan := 0;
  Result := iPan;
end;

// TSongListItem
//-----------------------------------------------------

constructor TSongListItem.Create(Collection: TCollection);
begin
  inherited;

  FPan := 0;
  FVolume := 100;
  FPosition := 0.0;

  FInitialized := False;

  //FTest := TTest.Create();
end;

//-----------------------------------------------------

destructor TSongListItem.Destroy;
begin
  CleanUpDirectShow;
  //FTest.Free;
  inherited;
end;

//-----------------------------------------------------

procedure TSongListItem.CleanUpDirectShow;
begin
  if FBasicAudio <> nil then FBasicAudio := nil;
  if FMediaPosition <> nil then FMediaPosition := nil;
  if FMediaControl <> nil then FMediaControl := nil;
  if FGraphBuilder <> nil then FGraphBuilder := nil;
end;

//-----------------------------------------------------

procedure TSongListItem.InitDirectShow;
begin
  if Failed(CoCreateInstance(CLSID_FilterGraph, nil, CLSCTX_INPROC, IID_IGraphBuilder, FGraphBuilder)) then
     raise Exception.Create('Failed to initialise DirectX Audio. DirectX 8 may not be installed');

  if Failed(CoCreateInstance(CLSID_FilterGraph, nil, CLSCTX_INPROC, IID_IGraphBuilder, FGraphBuilder)) then
     raise Exception.Create('Failed to initialise DirectShow. DirectX 8 may not be installed');
  if Failed(FGraphBuilder.QueryInterface(IID_IMediaControl, FMediaControl)) then
     raise Exception.Create('Failed to initialise DirectShow. DirectX 8 may not be installed');
  if Failed(FGraphBuilder.QueryInterface(IID_IMediaPosition, FMediaPosition)) then
     raise Exception.Create('Failed to initialise DirectShow. DirectX 8 may not be installed');
  if Failed(FGraphBuilder.QueryInterface(IID_IBasicAudio, FBasicAudio)) then
     raise Exception.Create('Failed to initialise DirectShow. DirectX 8 may not be installed');
end;

//-----------------------------------------------------

procedure TSongListItem.Init;
begin
  if not FLoaded then
  begin
    SetVolume(FVolume);
    SetPan(FPan);
    Load(FFilename);
  end;
  FInitialized := True;
end;

//-----------------------------------------------------

procedure TSongListItem.Load;
begin
  Load(FFileName);
end;

//-----------------------------------------------------

procedure TSongListItem.Load(const FileName: String);
begin
  if Filename = '' then Exit;

  CleanUpDirectShow;
  InitDirectShow;

  if Failed(FGraphBuilder.RenderFile(PWideChar(WideString(Filename)), nil)) then
  begin
    //raise Exception.Create('Error loading file: ' + FileName);
    Exit;
  end;

  FLoaded := True;
end;

//-----------------------------------------------------

procedure TSongListItem.SetPan(Value: Integer);
begin
  if Value > 100 then Value := 100;
  if Value < -100 then Value := -100;
  FPan := Value;

  if FBasicAudio <> nil then
     if Failed(FBasicAudio.put_Balance(DSPan(Value))) then
        raise Exception.Create('Error setting Pan');
end;

//-----------------------------------------------------

procedure TSongListItem.SetVolume(Value: Integer);
begin
  if Value > 100 then Value := 100;
  if Value < 0 then Value := 0;
  FVolume := Value;

  if FBasicAudio <> nil then
     if Failed(FBasicAudio.put_Volume(DSVolume(Value))) then
        //raise Exception.Create('Error setting volume');
end;

//-----------------------------------------------------

procedure TSongListItem.SetPosition(Value: Double);
begin
  if FMediaPosition <> nil then
     if Failed(FMediaPosition.put_CurrentPosition(Value)) then
        //raise Exception.Create('Error setting position');
end;

//-----------------------------------------------------

function TSongListItem.GetPosition: Double;
begin
  Result := 0.0;
  if FMediaPosition <> nil then
     if Failed(FMediaPosition.get_CurrentPosition(Result)) then
        raise Exception.Create('Error getting position');
end;

//-----------------------------------------------------

function TSongListItem.GetDuration: Double;
begin
  Result := 0.0;
  if FMediaPosition <> nil then
     if Failed(FMediaPosition.get_Duration(Result)) then
        //raise Exception.Create('Error getting duration');
end;

//-----------------------------------------------------

function TSongListItem.IsPlaying: Boolean; //TIsPlaying;
var
  refCurrent, refDuration: TREFTIME;
begin
  //Result := ipNotPlaying;
  Result := False;

  refCurrent := GetPosition;
  refDuration := GetDuration;
  if refCurrent < refDuration then Result := True;//Result := ipPlaying;
  //if FPaused then Result := ipPaused;
end;

//-----------------------------------------------------

procedure TSongListItem.LoopIt;
begin
  //if IsPlaying = ipNotPlaying then SetPosition(0.0);
  if not IsPlaying then SetPosition(0);
end;

//-----------------------------------------------------

procedure TSongListItem.Play;
begin
  if not FInitialized then Exit;

  if (FMediaControl <> nil) and (FMediaPosition <> nil) then
  begin
    SetPosition(0.0);
    if Failed(FMediaControl.Run) then
       raise Exception.Create('Error playing sound');
  end;
end;

//-----------------------------------------------------

function TSongListItem.GetDisplayName: String;
begin
  Result := FName;
end;

//-----------------------------------------------------

procedure TSongListItem.Stop;
begin
  if (FMediaControl <> nil) and (FMediaPosition <> nil) then
  begin
    if Failed(FMediaControl.Stop) then
       raise Exception.Create('Error stopping sound');
    SetPosition(0.0);
  end;
end;

//-----------------------------------------------------

procedure TSongListItem.Pause;
begin
  FPaused := not FPaused;
  if FPaused then
  begin
    if (FMediaControl <> nil) and (FMediaPosition <> nil) then
       if Failed(FMediaControl.Stop) then
          raise Exception.Create('Error pausing sound');
  end
  else
  begin
    if Failed(FMediaControl.Run) then
       raise Exception.Create('Error playing sound');
  end;
end;

// TSongListCollection
//-----------------------------------------------------

constructor TSongListCollection.Create(AOwner: TComponent);
begin
  inherited Create(AOwner, TSongListItem);
  if Failed(CoInitialize(nil)) then raise Exception.Create('Failed to initialise COM');
end;

//-----------------------------------------------------

destructor TSongListCollection.Destroy;
begin
  CoUninitialize;
  inherited;
end;

//-----------------------------------------------------

function TSongListCollection.GetItem(Index: Integer): TSongListItem;
begin
  Result := TSongListItem(inherited GetItem(Index));
end;

//-----------------------------------------------------

procedure TSongListCollection.SetItem(Index: Integer; Value: TSongListItem);
begin
  inherited SetItem(Index, Value);
end;

//-----------------------------------------------------

procedure TSongListCollection.Update(Item: TCollectionItem);
begin
  inherited Update(Item);
end;

//-----------------------------------------------------

function TSongListCollection.Add: TSongListItem;
begin
  Result := TSongListItem(inherited Add);
end;

//-----------------------------------------------------

function TSongListCollection.Insert(Index: Integer): TSongListItem;
begin
  Result := TSongListItem(inherited Insert(Index));
end;

//-----------------------------------------------------

function TSongListCollection.Find(AName: String): TSongListItem;
var
  i: Integer;
begin
  Result := nil;
  for i := 0 to Count - 1 do
  begin
    if AName = Items[i].FName then
    begin
      Result := Items[i];
      Exit;
    end;
  end;
end;

//-----------------------------------------------------

procedure TSongListCollection.Loop;
var
  i: Integer;
begin
  for i := 0 to Count - 1 do
  begin
    if Items[i].FLoop then
       Items[i].LoopIt;
  end;
end;

//-----------------------------------------------------

{type
  TSongListCollectionComponent = class(TComponent)
  private
    FList: TSongListCollection;
  public
    property List: TSongListCollection read FList write FList;
  end;

//-----------------------------------------------------

procedure TSongListCollection.SaveToFile(const FileName: String);
var
  Stream: TFileStream;
begin
  Stream := TFileStream.Create(FileName, fmCreate);
  try
    SaveToStream(Stream);
  finally
    Stream.Free;
  end;
end;

//-----------------------------------------------------

procedure TSongListCollection.SaveToStream(Stream: TStream);
var
  Component: TSongListCollectionComponent;
begin
  Component := TSongListCollectionComponent.Create(nil);
  try
    Component.FList := self;
    Stream.WriteComponentRes('OmegaSongListCollection', Component);
  finally
    Component.Free;
  end;
end;

//-----------------------------------------------------

procedure TSongListCollection.LoadFromFile(const FileName: String);
var
  Stream: TStream;
begin
  Stream := TFileStream.Create(FileName, fmOpenRead or fmShareDenyWrite);
  try
    LoadFromStream(Stream);
  finally
    Stream.Free;
  end;
end;

//-----------------------------------------------------

procedure TSongListCollection.LoadFromStream(Stream: TStream);
var
  Component: TSongListCollectionComponent;
begin
  Component := TSongListCollectionComponent.Create(nil);
  try
    Component.FList := self;
    Stream.ReadComponentRes(Component);
  finally
    Component.Free;
  end;
end;    }

// TOmegaMusic
//-----------------------------------------------------

constructor TOmegaMusic.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  FSongList := TSongListCollection.Create(self);
end;

//-----------------------------------------------------

destructor TOmegaMusic.Destroy;
begin
  FSongList.Free;
  inherited;
end;

//-----------------------------------------------------

procedure TOmegaMusic.SetSongList(const Value: TSongListCollection);
begin
  FSongList.Assign(Value);
end;

//-----------------------------------------------------

procedure TOmegaMusic.Init;
var
  i: Integer;
begin
  for i := 0 to FSongList.Count -1 do
  begin
    FSongList.Items[i].Init;
  end;
end;

//-----------------------------------------------------

procedure TOmegaMusic.Loop;
begin
  // it's easier to just write OmegaMusic1.Loop instead of OmegaMusic1.Songs.Loop
  FSongList.Loop;
end;

//-----------------------------------------------------

end.
